%%
%% %CopyrightBegin%
%%
%% SPDX-License-Identifier: Apache-2.0
%%
%% Copyright Ericsson AB 2020-2025. All Rights Reserved.
%%
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.
%%
%% %CopyrightEnd%
%%

%%
%% This file is included by the file "otp_internal.erl", which is
%% auto-generated by stdlib/scripts/update_deprecations
%%

-export([obsolete/3, obsolete_type/3, obsolete_callback/3]).

-type tag()     :: 'deprecated' | 'removed'. %% | 'experimental'.
-type mfas()    :: mfa() | {atom(), atom(), [byte()]} | string().
-type release() :: string().

-spec obsolete(module(), atom(), arity()) ->
        'no' | {tag(), string()} | {tag(), mfas(), release()}.

-spec obsolete_type(module(), atom(), arity()) ->
        'no' | {tag(), string()} | {tag(), mfas(), release()}.

-spec obsolete_callback(module(), atom(), arity()) ->
        'no' | {tag(), string()} | {tag(), mfas(), release()}.
