%%  Warning: Do not edit this file.
%%  Auto-generated by 'beam_makeops'.

-define(tag_u, 0).
-define(tag_i, 1).
-define(tag_a, 2).
-define(tag_x, 3).
-define(tag_y, 4).
-define(tag_f, 5).
-define(tag_h, 6).
-define(tag_z, 7).

-define(BEAMop_label, 1).
-define(BEAMop_ARGS_label, 1).
-define(BEAMop_func_info, 2).
-define(BEAMop_ARGS_func_info, 3).
-define(BEAMop_int_code_end, 3).
-define(BEAMop_ARGS_int_code_end, 0).
-define(BEAMop_call, 4).
-define(BEAMop_ARGS_call, 2).
-define(BEAMop_call_last, 5).
-define(BEAMop_ARGS_call_last, 3).
-define(BEAMop_call_only, 6).
-define(BEAMop_ARGS_call_only, 2).
-define(BEAMop_call_ext, 7).
-define(BEAMop_ARGS_call_ext, 2).
-define(BEAMop_call_ext_last, 8).
-define(BEAMop_ARGS_call_ext_last, 3).
-define(BEAMop_bif0, 9).
-define(BEAMop_ARGS_bif0, 2).
-define(BEAMop_bif1, 10).
-define(BEAMop_ARGS_bif1, 4).
-define(BEAMop_bif2, 11).
-define(BEAMop_ARGS_bif2, 5).
-define(BEAMop_allocate, 12).
-define(BEAMop_ARGS_allocate, 2).
-define(BEAMop_allocate_heap, 13).
-define(BEAMop_ARGS_allocate_heap, 3).
-define(BEAMop_allocate_zero, 14).
-define(BEAMop_ARGS_allocate_zero, 2).
-define(BEAMop_allocate_heap_zero, 15).
-define(BEAMop_ARGS_allocate_heap_zero, 3).
-define(BEAMop_test_heap, 16).
-define(BEAMop_ARGS_test_heap, 2).
-define(BEAMop_init, 17).
-define(BEAMop_ARGS_init, 1).
-define(BEAMop_deallocate, 18).
-define(BEAMop_ARGS_deallocate, 1).
-define(BEAMop_return, 19).
-define(BEAMop_ARGS_return, 0).
-define(BEAMop_send, 20).
-define(BEAMop_ARGS_send, 0).
-define(BEAMop_remove_message, 21).
-define(BEAMop_ARGS_remove_message, 0).
-define(BEAMop_timeout, 22).
-define(BEAMop_ARGS_timeout, 0).
-define(BEAMop_loop_rec, 23).
-define(BEAMop_ARGS_loop_rec, 2).
-define(BEAMop_loop_rec_end, 24).
-define(BEAMop_ARGS_loop_rec_end, 1).
-define(BEAMop_wait, 25).
-define(BEAMop_ARGS_wait, 1).
-define(BEAMop_wait_timeout, 26).
-define(BEAMop_ARGS_wait_timeout, 2).
-define(BEAMop_m_plus, 27).
-define(BEAMop_ARGS_m_plus, 4).
-define(BEAMop_m_minus, 28).
-define(BEAMop_ARGS_m_minus, 4).
-define(BEAMop_m_times, 29).
-define(BEAMop_ARGS_m_times, 4).
-define(BEAMop_m_div, 30).
-define(BEAMop_ARGS_m_div, 4).
-define(BEAMop_int_div, 31).
-define(BEAMop_ARGS_int_div, 4).
-define(BEAMop_int_rem, 32).
-define(BEAMop_ARGS_int_rem, 4).
-define(BEAMop_int_band, 33).
-define(BEAMop_ARGS_int_band, 4).
-define(BEAMop_int_bor, 34).
-define(BEAMop_ARGS_int_bor, 4).
-define(BEAMop_int_bxor, 35).
-define(BEAMop_ARGS_int_bxor, 4).
-define(BEAMop_int_bsl, 36).
-define(BEAMop_ARGS_int_bsl, 4).
-define(BEAMop_int_bsr, 37).
-define(BEAMop_ARGS_int_bsr, 4).
-define(BEAMop_int_bnot, 38).
-define(BEAMop_ARGS_int_bnot, 3).
-define(BEAMop_is_lt, 39).
-define(BEAMop_ARGS_is_lt, 3).
-define(BEAMop_is_ge, 40).
-define(BEAMop_ARGS_is_ge, 3).
-define(BEAMop_is_eq, 41).
-define(BEAMop_ARGS_is_eq, 3).
-define(BEAMop_is_ne, 42).
-define(BEAMop_ARGS_is_ne, 3).
-define(BEAMop_is_eq_exact, 43).
-define(BEAMop_ARGS_is_eq_exact, 3).
-define(BEAMop_is_ne_exact, 44).
-define(BEAMop_ARGS_is_ne_exact, 3).
-define(BEAMop_is_integer, 45).
-define(BEAMop_ARGS_is_integer, 2).
-define(BEAMop_is_float, 46).
-define(BEAMop_ARGS_is_float, 2).
-define(BEAMop_is_number, 47).
-define(BEAMop_ARGS_is_number, 2).
-define(BEAMop_is_atom, 48).
-define(BEAMop_ARGS_is_atom, 2).
-define(BEAMop_is_pid, 49).
-define(BEAMop_ARGS_is_pid, 2).
-define(BEAMop_is_reference, 50).
-define(BEAMop_ARGS_is_reference, 2).
-define(BEAMop_is_port, 51).
-define(BEAMop_ARGS_is_port, 2).
-define(BEAMop_is_nil, 52).
-define(BEAMop_ARGS_is_nil, 2).
-define(BEAMop_is_binary, 53).
-define(BEAMop_ARGS_is_binary, 2).
-define(BEAMop_is_constant, 54).
-define(BEAMop_ARGS_is_constant, 2).
-define(BEAMop_is_list, 55).
-define(BEAMop_ARGS_is_list, 2).
-define(BEAMop_is_nonempty_list, 56).
-define(BEAMop_ARGS_is_nonempty_list, 2).
-define(BEAMop_is_tuple, 57).
-define(BEAMop_ARGS_is_tuple, 2).
-define(BEAMop_test_arity, 58).
-define(BEAMop_ARGS_test_arity, 3).
-define(BEAMop_select_val, 59).
-define(BEAMop_ARGS_select_val, 3).
-define(BEAMop_select_tuple_arity, 60).
-define(BEAMop_ARGS_select_tuple_arity, 3).
-define(BEAMop_jump, 61).
-define(BEAMop_ARGS_jump, 1).
-define(BEAMop_catch, 62).
-define(BEAMop_ARGS_catch, 2).
-define(BEAMop_catch_end, 63).
-define(BEAMop_ARGS_catch_end, 1).
-define(BEAMop_move, 64).
-define(BEAMop_ARGS_move, 2).
-define(BEAMop_get_list, 65).
-define(BEAMop_ARGS_get_list, 3).
-define(BEAMop_get_tuple_element, 66).
-define(BEAMop_ARGS_get_tuple_element, 3).
-define(BEAMop_set_tuple_element, 67).
-define(BEAMop_ARGS_set_tuple_element, 3).
-define(BEAMop_put_string, 68).
-define(BEAMop_ARGS_put_string, 3).
-define(BEAMop_put_list, 69).
-define(BEAMop_ARGS_put_list, 3).
-define(BEAMop_put_tuple, 70).
-define(BEAMop_ARGS_put_tuple, 2).
-define(BEAMop_put, 71).
-define(BEAMop_ARGS_put, 1).
-define(BEAMop_badmatch, 72).
-define(BEAMop_ARGS_badmatch, 1).
-define(BEAMop_if_end, 73).
-define(BEAMop_ARGS_if_end, 0).
-define(BEAMop_case_end, 74).
-define(BEAMop_ARGS_case_end, 1).
-define(BEAMop_call_fun, 75).
-define(BEAMop_ARGS_call_fun, 1).
-define(BEAMop_make_fun, 76).
-define(BEAMop_ARGS_make_fun, 3).
-define(BEAMop_is_function, 77).
-define(BEAMop_ARGS_is_function, 2).
-define(BEAMop_call_ext_only, 78).
-define(BEAMop_ARGS_call_ext_only, 2).
-define(BEAMop_bs_start_match, 79).
-define(BEAMop_ARGS_bs_start_match, 2).
-define(BEAMop_bs_get_integer, 80).
-define(BEAMop_ARGS_bs_get_integer, 5).
-define(BEAMop_bs_get_float, 81).
-define(BEAMop_ARGS_bs_get_float, 5).
-define(BEAMop_bs_get_binary, 82).
-define(BEAMop_ARGS_bs_get_binary, 5).
-define(BEAMop_bs_skip_bits, 83).
-define(BEAMop_ARGS_bs_skip_bits, 4).
-define(BEAMop_bs_test_tail, 84).
-define(BEAMop_ARGS_bs_test_tail, 2).
-define(BEAMop_bs_save, 85).
-define(BEAMop_ARGS_bs_save, 1).
-define(BEAMop_bs_restore, 86).
-define(BEAMop_ARGS_bs_restore, 1).
-define(BEAMop_bs_init, 87).
-define(BEAMop_ARGS_bs_init, 2).
-define(BEAMop_bs_final, 88).
-define(BEAMop_ARGS_bs_final, 2).
-define(BEAMop_bs_put_integer, 89).
-define(BEAMop_ARGS_bs_put_integer, 5).
-define(BEAMop_bs_put_binary, 90).
-define(BEAMop_ARGS_bs_put_binary, 5).
-define(BEAMop_bs_put_float, 91).
-define(BEAMop_ARGS_bs_put_float, 5).
-define(BEAMop_bs_put_string, 92).
-define(BEAMop_ARGS_bs_put_string, 2).
-define(BEAMop_bs_need_buf, 93).
-define(BEAMop_ARGS_bs_need_buf, 1).
-define(BEAMop_fclearerror, 94).
-define(BEAMop_ARGS_fclearerror, 0).
-define(BEAMop_fcheckerror, 95).
-define(BEAMop_ARGS_fcheckerror, 1).
-define(BEAMop_fmove, 96).
-define(BEAMop_ARGS_fmove, 2).
-define(BEAMop_fconv, 97).
-define(BEAMop_ARGS_fconv, 2).
-define(BEAMop_fadd, 98).
-define(BEAMop_ARGS_fadd, 4).
-define(BEAMop_fsub, 99).
-define(BEAMop_ARGS_fsub, 4).
-define(BEAMop_fmul, 100).
-define(BEAMop_ARGS_fmul, 4).
-define(BEAMop_fdiv, 101).
-define(BEAMop_ARGS_fdiv, 4).
-define(BEAMop_fnegate, 102).
-define(BEAMop_ARGS_fnegate, 3).
-define(BEAMop_make_fun2, 103).
-define(BEAMop_ARGS_make_fun2, 1).
-define(BEAMop_try, 104).
-define(BEAMop_ARGS_try, 2).
-define(BEAMop_try_end, 105).
-define(BEAMop_ARGS_try_end, 1).
-define(BEAMop_try_case, 106).
-define(BEAMop_ARGS_try_case, 1).
-define(BEAMop_try_case_end, 107).
-define(BEAMop_ARGS_try_case_end, 1).
-define(BEAMop_raise, 108).
-define(BEAMop_ARGS_raise, 2).
-define(BEAMop_bs_init2, 109).
-define(BEAMop_ARGS_bs_init2, 6).
-define(BEAMop_bs_bits_to_bytes, 110).
-define(BEAMop_ARGS_bs_bits_to_bytes, 3).
-define(BEAMop_bs_add, 111).
-define(BEAMop_ARGS_bs_add, 5).
-define(BEAMop_apply, 112).
-define(BEAMop_ARGS_apply, 1).
-define(BEAMop_apply_last, 113).
-define(BEAMop_ARGS_apply_last, 2).
-define(BEAMop_is_boolean, 114).
-define(BEAMop_ARGS_is_boolean, 2).
-define(BEAMop_is_function2, 115).
-define(BEAMop_ARGS_is_function2, 3).
-define(BEAMop_bs_start_match2, 116).
-define(BEAMop_ARGS_bs_start_match2, 5).
-define(BEAMop_bs_get_integer2, 117).
-define(BEAMop_ARGS_bs_get_integer2, 7).
-define(BEAMop_bs_get_float2, 118).
-define(BEAMop_ARGS_bs_get_float2, 7).
-define(BEAMop_bs_get_binary2, 119).
-define(BEAMop_ARGS_bs_get_binary2, 7).
-define(BEAMop_bs_skip_bits2, 120).
-define(BEAMop_ARGS_bs_skip_bits2, 5).
-define(BEAMop_bs_test_tail2, 121).
-define(BEAMop_ARGS_bs_test_tail2, 3).
-define(BEAMop_bs_save2, 122).
-define(BEAMop_ARGS_bs_save2, 2).
-define(BEAMop_bs_restore2, 123).
-define(BEAMop_ARGS_bs_restore2, 2).
-define(BEAMop_gc_bif1, 124).
-define(BEAMop_ARGS_gc_bif1, 5).
-define(BEAMop_gc_bif2, 125).
-define(BEAMop_ARGS_gc_bif2, 6).
-define(BEAMop_bs_final2, 126).
-define(BEAMop_ARGS_bs_final2, 2).
-define(BEAMop_bs_bits_to_bytes2, 127).
-define(BEAMop_ARGS_bs_bits_to_bytes2, 2).
-define(BEAMop_put_literal, 128).
-define(BEAMop_ARGS_put_literal, 2).
-define(BEAMop_is_bitstr, 129).
-define(BEAMop_ARGS_is_bitstr, 2).
-define(BEAMop_bs_context_to_binary, 130).
-define(BEAMop_ARGS_bs_context_to_binary, 1).
-define(BEAMop_bs_test_unit, 131).
-define(BEAMop_ARGS_bs_test_unit, 3).
-define(BEAMop_bs_match_string, 132).
-define(BEAMop_ARGS_bs_match_string, 4).
-define(BEAMop_bs_init_writable, 133).
-define(BEAMop_ARGS_bs_init_writable, 0).
-define(BEAMop_bs_append, 134).
-define(BEAMop_ARGS_bs_append, 8).
-define(BEAMop_bs_private_append, 135).
-define(BEAMop_ARGS_bs_private_append, 6).
-define(BEAMop_trim, 136).
-define(BEAMop_ARGS_trim, 2).
-define(BEAMop_bs_init_bits, 137).
-define(BEAMop_ARGS_bs_init_bits, 6).
-define(BEAMop_bs_get_utf8, 138).
-define(BEAMop_ARGS_bs_get_utf8, 5).
-define(BEAMop_bs_skip_utf8, 139).
-define(BEAMop_ARGS_bs_skip_utf8, 4).
-define(BEAMop_bs_get_utf16, 140).
-define(BEAMop_ARGS_bs_get_utf16, 5).
-define(BEAMop_bs_skip_utf16, 141).
-define(BEAMop_ARGS_bs_skip_utf16, 4).
-define(BEAMop_bs_get_utf32, 142).
-define(BEAMop_ARGS_bs_get_utf32, 5).
-define(BEAMop_bs_skip_utf32, 143).
-define(BEAMop_ARGS_bs_skip_utf32, 4).
-define(BEAMop_bs_utf8_size, 144).
-define(BEAMop_ARGS_bs_utf8_size, 3).
-define(BEAMop_bs_put_utf8, 145).
-define(BEAMop_ARGS_bs_put_utf8, 3).
-define(BEAMop_bs_utf16_size, 146).
-define(BEAMop_ARGS_bs_utf16_size, 3).
-define(BEAMop_bs_put_utf16, 147).
-define(BEAMop_ARGS_bs_put_utf16, 3).
-define(BEAMop_bs_put_utf32, 148).
-define(BEAMop_ARGS_bs_put_utf32, 3).
-define(BEAMop_on_load, 149).
-define(BEAMop_ARGS_on_load, 0).
-define(BEAMop_recv_mark, 150).
-define(BEAMop_ARGS_recv_mark, 1).
-define(BEAMop_recv_set, 151).
-define(BEAMop_ARGS_recv_set, 1).
-define(BEAMop_gc_bif3, 152).
-define(BEAMop_ARGS_gc_bif3, 7).
-define(BEAMop_line, 153).
-define(BEAMop_ARGS_line, 1).
-define(BEAMop_put_map_assoc, 154).
-define(BEAMop_ARGS_put_map_assoc, 5).
-define(BEAMop_put_map_exact, 155).
-define(BEAMop_ARGS_put_map_exact, 5).
-define(BEAMop_is_map, 156).
-define(BEAMop_ARGS_is_map, 2).
-define(BEAMop_has_map_fields, 157).
-define(BEAMop_ARGS_has_map_fields, 3).
-define(BEAMop_get_map_elements, 158).
-define(BEAMop_ARGS_get_map_elements, 3).
-define(BEAMop_is_tagged_tuple, 159).
-define(BEAMop_ARGS_is_tagged_tuple, 4).
-define(BEAMop_build_stacktrace, 160).
-define(BEAMop_ARGS_build_stacktrace, 0).
-define(BEAMop_raw_raise, 161).
-define(BEAMop_ARGS_raw_raise, 0).
-define(BEAMop_get_hd, 162).
-define(BEAMop_ARGS_get_hd, 2).
-define(BEAMop_get_tl, 163).
-define(BEAMop_ARGS_get_tl, 2).
-define(BEAMop_put_tuple2, 164).
-define(BEAMop_ARGS_put_tuple2, 2).
-define(BEAMop_bs_get_tail, 165).
-define(BEAMop_ARGS_bs_get_tail, 3).
-define(BEAMop_bs_start_match3, 166).
-define(BEAMop_ARGS_bs_start_match3, 4).
-define(BEAMop_bs_get_position, 167).
-define(BEAMop_ARGS_bs_get_position, 3).
-define(BEAMop_bs_set_position, 168).
-define(BEAMop_ARGS_bs_set_position, 2).
-define(BEAMop_swap, 169).
-define(BEAMop_ARGS_swap, 2).
-define(BEAMop_bs_start_match4, 170).
-define(BEAMop_ARGS_bs_start_match4, 4).
-define(BEAMop_make_fun3, 171).
-define(BEAMop_ARGS_make_fun3, 3).
-define(BEAMop_init_yregs, 172).
-define(BEAMop_ARGS_init_yregs, 1).
-define(BEAMop_recv_marker_bind, 173).
-define(BEAMop_ARGS_recv_marker_bind, 2).
-define(BEAMop_recv_marker_clear, 174).
-define(BEAMop_ARGS_recv_marker_clear, 1).
-define(BEAMop_recv_marker_reserve, 175).
-define(BEAMop_ARGS_recv_marker_reserve, 1).
-define(BEAMop_recv_marker_use, 176).
-define(BEAMop_ARGS_recv_marker_use, 1).
-define(BEAMop_bs_create_bin, 177).
-define(BEAMop_ARGS_bs_create_bin, 6).
-define(BEAMop_call_fun2, 178).
-define(BEAMop_ARGS_call_fun2, 3).
-define(BEAMop_nif_start, 179).
-define(BEAMop_ARGS_nif_start, 0).
-define(BEAMop_badrecord, 180).
-define(BEAMop_ARGS_badrecord, 1).
-define(BEAMop_update_record, 181).
-define(BEAMop_ARGS_update_record, 5).
-define(BEAMop_bs_match, 182).
-define(BEAMop_ARGS_bs_match, 3).
-define(BEAMop_executable_line, 183).
-define(BEAMop_ARGS_executable_line, 2).
-define(BEAMop_debug_line, 184).
-define(BEAMop_ARGS_debug_line, 4).
-define(BEAMop_bif3, 185).
-define(BEAMop_ARGS_bif3, 6).
-define(BEAMop_is_any_native_record, 186).
-define(BEAMop_ARGS_is_any_native_record, 2).
-define(BEAMop_is_native_record, 187).
-define(BEAMop_ARGS_is_native_record, 4).
-define(BEAMop_get_record_elements, 188).
-define(BEAMop_ARGS_get_record_elements, 3).
-define(BEAMop_put_record, 189).
-define(BEAMop_ARGS_put_record, 6).
-define(BEAMop_is_record_accessible, 190).
-define(BEAMop_ARGS_is_record_accessible, 3).
-define(BEAMop_get_record_field, 191).
-define(BEAMop_ARGS_get_record_field, 5).
